@@ -234,11 +234,13 @@ void ssl_io_suck(request_rec *r, SSL *ssl)
 
             /* suck trailing data (usually CR LF) which 
                is still in the Apache BUFF layer */
+            ap_hard_timeout("SSL I/O request trailing data pre-sucking", r);
             while (ap_bpeekc(r->connection->client) != EOF) {
                 c = ap_bgetc(r->connection->client);
                 ssl_io_suck_record(r, &c, 1);
                 sucked++;
             }
+            ap_kill_timeout(r);
 
             ssl_log(r->server, SSL_LOG_TRACE, 
                     "I/O: sucked %d bytes of input data from SSL/TLS I/O layer "
@@ -443,10 +445,6 @@ static int ssl_io_hook_writev(BUFF *fb, const struct iovec *iov, int iovcnt)
 
 #ifdef WIN32
 
-/* these two functions are exported from buff.c under WIN32 */
-API_EXPORT(int) sendwithtimeout(int sock, const char *buf, int len, int flags);
-API_EXPORT(int) recvwithtimeout(int sock, char *buf, int len, int flags);
-
 /* and the prototypes for our SSL_xxx variants */
 static int SSL_sendwithtimeout(BUFF *fb, const char *buf, int len);
 static int SSL_recvwithtimeout(BUFF *fb, char *buf, int len);
@@ -459,7 +457,7 @@ static int ssl_io_hook_recvwithtimeout(BUFF *fb, char *buf, int len)
     if ((ssl = ap_ctx_get(fb->ctx, "ssl")) != NULL)
         rc = SSL_recvwithtimeout(fb, buf, len);
     else
-        rc = recvwithtimeout(fb->fd, buf, len, 0);
+        rc = ap_recvwithtimeout(fb->fd, buf, len, 0);
     return rc;
 }
 
@@ -471,7 +469,7 @@ static int ssl_io_hook_sendwithtimeout(BUFF *fb, const char *buf, int len)
     if ((ssl = ap_ctx_get(fb->ctx, "ssl")) != NULL)
         rc = SSL_sendwithtimeout(fb, buf, len);
     else
-        rc = sendwithtimeout(fb->fd, buf, len, 0);
+        rc = ap_sendwithtimeout(fb->fd, buf, len, 0);
     return rc;
 }
 
